import sys, socket, gc

# gc.set_threshold(100, 10, 10)
gc.set_debug(gc.DEBUG_LEAK)

from peak.api import binding, events, PropertyName
from peak.running import commands, interfaces, daemons


class ComponentBase(binding.Component):
    eventLoop = binding.Obtain(events.IEventLoop)
    log= binding.Obtain(PropertyName("peak.logs.App"))

    terminated = False

    def terminate(self):
        self.terminated = True

class CommandBase(commands.AbstractCommand, ComponentBase):
    mainLoop = binding.Obtain(interfaces.IMainLoop)

    def _run(self):
        return self.mainLoop.run()

class Alarm(ComponentBase, daemons.AdaptiveTask):
    callBack = binding.Require("Procedure to call")

    def __call__(self):
        if self.terminated:
            raise daemons.exceptions.StopRunning
        else:
            return super(Alarm,self).__call__()

    def getWork(self):
        """Return True"""
        return True

    def doWork(self, job):
        self.callBack()
        return True


class SocketMAC(CommandBase):

    socket = binding.Require()

    PING_INTERVAL = 2
    RECEIVE_TIMEOUT = 5

    PING_MESSAGE = "\x00"
    PONG_MESSAGE = "\x01"

    _pinger = binding.Make(
        lambda self: Alarm(runEvery=self.PING_INTERVAL,
            callBack=self.sendPingMsg, parentComponent=self),
        uponAssembly=True)

    exceptional_evt = binding.Make(
        lambda self: self.eventLoop.exceptional(self.socket))
    readable_evt = binding.Make(
        lambda self: self.eventLoop.readable(self.socket))
    writable_evt = binding.Make(
        lambda self: self.eventLoop.writable(self.socket))
    timeout_evt = None

    _outputQueue = ""
    _outputQueueOOB = ""

    ##

    def _send(self, msg, oobFlag=False):
        assert not self.terminated
        if oobFlag:
            self._outputQueueOOB += msg
        else:
            self._outputQueue = msg

    def _sendToSocket(self, msg, oobFlag=False):
        _flags = (0, socket.MSG_OOB)[oobFlag]
        try:
            return self.socket.send(msg, _flags)
        except:
            self.log.exception("Error sending to socket (oobFlag=%s)", oobFlag)
            self.terminate()
            return 0

    def _recvFromSocket(self, oobFlag=False):
        try:
            _rv = self.socket.recv(1024, (0, socket.MSG_OOB)[oobFlag])
        except:
            self.log.exception("Error receiving from socket (oobFlag=%s)",
                oobFlag)
            self.terminate()
            return ""
        if not _rv:
            self.terminate()
        else:
            self.timeout_evt = None
            print gc.collect()
        return _rv

    ##

    def terminate(self):
        for _attr in ("readable", "writable", "exceptional"):
            self._delBinding(_attr + "_evt")
        self._pinger.terminate()
        self._delBinding("_pinger")
        self.socket.close()
        super(SocketMAC,self).terminate()

    def evtLoop(self):
        while not self.terminated:
            _evt = self._getEvent()
            yield _evt
            _evt = events.resume()
            self._handleEvent(_evt)
        self.log.notice("Terminated")
        self.mainLoop.exitWith(0)
    evtLoop = binding.Make(events.taskFactory(evtLoop), uponAssembly=True)

    def _getEvent(self):
        _evts = [self.readable_evt, self.exceptional_evt]
        if self._outputQueue or self._outputQueueOOB:
            _evts.append(self.writable_evt)
        if not self.timeout_evt:
            self.timeout_evt = self.eventLoop.sleep(self.RECEIVE_TIMEOUT)
        _evts.append(self.timeout_evt)
        return events.AnyOf(*_evts)

    def _handleEvent(self, event):
        _evt = event[0]
        if _evt is self.exceptional_evt:
            self.doReadExceptional()
        elif _evt is self.readable_evt:
            self.doRead()
        elif _evt is self.writable_evt:
            self.doWrite()
        elif self.timeout_evt and _evt is self.timeout_evt:
            self.handleTimeout()
        else:
            self.log.warning("Unknown event %r", event)
    ##

    def sendPingMsg(self):
        self._send(self.PING_MESSAGE, True)
        self.log.info("Send PING message")
        print gc.collect()
        print len(gc.get_objects())

    def doRead(self):
        _data = self._recvFromSocket()
        if _data: self.log.notice("Received %r from the socket", _data)

    def doWrite(self):
        if self._outputQueueOOB:
            _sent = self._sendToSocket(self._outputQueueOOB[0], True)
            if _sent:
                self._outputQueueOOB = self._outputQueueOOB[_sent:]
        else:
            _sent = self._sendToSocket(self._outputQueue)
            if _sent:
                self._outputQueue = self._outputQueue[_sent:]

    def doReadExceptional(self):
        _data = self._recvFromSocket(True)
        if not _data:
            return
        for _ch in _data:
            if _ch == self.PING_MESSAGE:
                self._send(self.PONG_MESSAGE, True)
            elif _ch != self.PONG_MESSAGE:
                self.log.warning("Unknown character %r", _ch)

    def handleTimeout(self):
        self.log.error("No data received from the client within %i seconds",
            self.RECEIVE_TIMEOUT)
        self.terminate()

class App(binding.Configurable, CommandBase):
    serverPort = binding.Obtain(PropertyName("app.serverPort"))
    interface = "127.0.0.1"

    def _run(self):
        _serverSocket = socket.socket()
        _bindTo = (self.interface, self.serverPort)
        _serverSocket.bind(_bindTo)
        _serverSocket.listen(1)
        self.log.notice("Waiting for connection (%s:%i)" % _bindTo)
        _clientSocket, _addrInfo = _serverSocket.accept()
        self.log.notice("Connection from %s:%i accepted" % _addrInfo)
        self.getSubcommand(SocketMAC, socket=_clientSocket)
        return super(App,self)._run()

if __name__ == "__main__":
    commands.runMain(App)

# vim: set sts=4 sw=4 et :
